# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from qat.purr.compiler.devices import Qubit, Resonator
from qat.purr.compiler.waveforms import AbstractWaveform
from qat.purr.integrations.qasm import (
    extern_port_name,
    get_frame_mappings,
    get_port_mappings,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class FeatureMetadata:
    """Exposed metadata for various languages, integrations and features."""

    def to_json_dict(self):
        """
        Turn object into a JSON-amenable dictionary to return from a web service. All
        returned dictionaries should be able to be merged together without additional
        effort, so make sure a root node with the name of the feature is available.
        """
        pass


class Scale(Enum):
    """
    SI units of frequency
    """

    NANO = "ns"
    MICRO = "us"
    MILLI = "ms"
    DEFAULT = ""
    KILO = "k"
    MEGA = "M"
    GIGA = "G"
    TERA = "T"


class Unit(Enum):
    """
    Physical SI units.
    """

    TIME = "s"
    FREQUENCY = "Hz"


# TODO: You don't want something lke this, just have two different objects for
#   time/frequency.
@dataclass(frozen=True)
class Quantity:
    amount: float
    unit: Unit
    scale: Scale

    def __str__(self):
        return f"{self.amount} {self.scale.value}{self.unit.value}"


@dataclass(frozen=True)
class Constraints:
    pulse_control_contraints: str = None
    max_scale: float = 1
    max_waveform_amplitude: float = 1
    awg_frequency_bandwidth: str = Quantity(1, Unit.FREQUENCY, Scale.DEFAULT)
    max_pulse_duration: str = Quantity(120, Unit.TIME, Scale.MICRO)
    min_pulse_duration: str = Quantity(8, Unit.TIME, Scale.NANO)
    min_pulse_time: str = Quantity(8, Unit.TIME, Scale.NANO)


class OpenPulseFeatures(FeatureMetadata):
    def __init__(self):
        self.ports = dict()
        self.frames = dict()
        self.waveforms = dict()
        self.constraints: Optional[Constraints] = None

    def for_hardware(self, hardware):
        def _find_qubit(qubit) -> Qubit:
            if isinstance(qubit, Resonator):
                for qb in hardware.qubits:
                    if qb.measure_device == qubit:
                        qubit = qb

            if isinstance(qubit, Resonator):
                raise ValueError(f"{str(qubit)} was unable to be matched to a qubit.")

            return qubit

        for frame_name, channel_view in get_frame_mappings(hardware).items():
            frame = channel_view.pulse_channel
            qubit = _find_qubit(hardware.get_devices_from_pulse_channel(frame.full_id())[0])
            qubits = [qubit.id]
            qubits.extend(
                [
                    qubit.id
                    for qubit in channel_view.auxiliary_devices
                    if isinstance(qubit, Qubit)
                ]
            )
            self.frames[frame_name] = dict(
                qubits=qubits,
                port_id=extern_port_name(frame.physical_channel),
                frequency=frame.frequency,
                bandwidth_centre=frame.baseband_frequency,
                phase=0.0,
            )

        for port_name, port in get_port_mappings(hardware).items():
            self.ports[port_name] = dict(
                direction="two-way" if port.acquire_allowed else "one-way",
                type="port_type_1",
                associated_qubits=[
                    _find_qubit(qb).index
                    for qb in hardware.quantum_devices.values()
                    if qb.physical_channel == port
                ],
            )

        self.waveforms = {
            key: vars(value.waveform_definition)
            for key, value in AbstractWaveform.actual_waveforms.items()
        }

        self.constraints = Constraints()

    def to_json_dict(self):
        return {
            "open_pulse": {
                "version": "0",
                "ports": self.ports,
                "frames": self.frames,
                "waveforms": self.waveforms,
                "constraints": self.constraints,
            }
        }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import warnings

warnings.simplefilter("always", DeprecationWarning)
warnings.warn(
    "module 'qat.purr.compiler.config' is deprecated, please use "
    "'compiler_config.serialiser instead.'",
    DeprecationWarning,
    stacklevel=2,
)

from compiler_config.config import *  # fmt: skip



import abc
from queue import Queue
from threading import Event


class InterruptError(RuntimeError): ...


class Interrupt(abc.ABC):
    @abc.abstractmethod
    def if_triggered(self, metadata={}, throw=False):
        """
        Capture context if triggered
        """
        ...

    @abc.abstractmethod
    def trigger(self):
        """
        Set the triggered status
        """
        ...


class NullInterrupt(Interrupt):
    def if_triggered(self, metadata={}, throw=False): ...

    def trigger(self): ...


class BasicInterrupt(Interrupt):
    """
    Interrupt supporting multi-threading and multi-process concurrency
    """

    def __init__(self, event=Event(), queue=Queue()):
        self._event = event
        self._queue = queue
        self._metadata = {}

    def if_triggered(self, metadata={}, throw=False):
        self._metadata.update(metadata)
        if self._event.is_set():
            response = {"interrupted": True, "metadata": self._metadata}
            self._queue.put(response)
            if throw:
                raise InterruptError(f"Interrupt triggered, {self._metadata}")

    def trigger(self):
        self._event.set()

    @property
    def event(self):
        return self._event

    @property
    def queue(self):
        return self._queue



from numbers import Number
from typing import List

import numpy as np

from qat import qatconfig
from qat.ir.pass_base import ValidationPass
from qat.ir.result_base import ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import MaxPulseLength, PulseChannel, Qubit
from qat.purr.compiler.execution import QuantumExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    CustomPulse,
    PostProcessing,
    ProcessAxis,
    Pulse,
    Sweep,
    Variable,
)


class InstructionValidation(ValidationPass):
    """
    Extracted from QuantumExecutionEngine.validate()
    """

    def run(self, builder: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        engine = next((a for a in args if isinstance(a, QuantumExecutionEngine)), None)

        if not engine:
            engine = kwargs.get("engine", None)

        if not engine or not isinstance(engine, QuantumExecutionEngine):
            raise ValueError(
                f"Expected to find an instance of {QuantumExecutionEngine} in arguments list, but got {engine} instead"
            )

        instruction_length = len(builder.instructions)
        if instruction_length > engine.max_instruction_len:
            raise ValueError(
                f"Program too large to be run in a single block on current hardware. "
                f"{instruction_length} instructions."
            )

        for inst in builder.instructions:
            if isinstance(inst, Acquire) and not inst.channel.acquire_allowed:
                raise ValueError(
                    f"Cannot perform an acquire on the physical channel with id "
                    f"{inst.channel.physical_channel}"
                )
            if isinstance(inst, (Pulse, CustomPulse)):
                duration = inst.duration
                if isinstance(duration, Number) and duration > MaxPulseLength:
                    if (
                        not qatconfig.DISABLE_PULSE_DURATION_LIMITS
                    ):  # Do not throw error if we specifically disabled the limit checks.
                        # TODO: Add a lower bound for the pulse duration limits as well in a later PR,
                        # which is specific to each hardware model and can be stored as a member variables there.
                        raise ValueError(
                            f"Max Waveform width is {MaxPulseLength} s "
                            f"given: {inst.duration} s"
                        )
                elif isinstance(duration, Variable):
                    values = next(
                        iter(
                            [
                                sw.variables[duration.name]
                                for sw in builder.instructions
                                if isinstance(sw, Sweep)
                                and duration.name in sw.variables.keys()
                            ]
                        )
                    )
                    if np.max(values) > MaxPulseLength:
                        if not qatconfig.DISABLE_PULSE_DURATION_LIMITS:
                            raise ValueError(
                                f"Max Waveform width is {MaxPulseLength} s "
                                f"given: {values} s"
                            )


class ReadoutValidation(ValidationPass):
    """
    Extracted from LiveDeviceEngine.optimize()
    """

    def run(self, builder: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        model = next((a for a in args if isinstance(a, QuantumHardwareModel)), None)

        if not model:
            model = kwargs.get("model", None)

        if not model or not isinstance(model, QuantumHardwareModel):
            raise ValueError(
                f"Expected to find an instance of {QuantumHardwareModel} in arguments list, but got {model} instead"
            )

        consumed_qubits: List[str] = []
        for inst in builder.instructions:
            if isinstance(inst, PostProcessing):
                if (
                    inst.acquire.mode == AcquireMode.SCOPE
                    and ProcessAxis.SEQUENCE in inst.axes
                ):
                    raise ValueError(
                        "Invalid post-processing! Post-processing over SEQUENCE is "
                        "not possible after the result is returned from hardware "
                        "in SCOPE mode!"
                    )
                elif (
                    inst.acquire.mode == AcquireMode.INTEGRATOR
                    and ProcessAxis.TIME in inst.axes
                ):
                    raise ValueError(
                        "Invalid post-processing! Post-processing over TIME is not "
                        "possible after the result is returned from hardware in "
                        "INTEGRATOR mode!"
                    )
                elif inst.acquire.mode == AcquireMode.RAW:
                    raise ValueError(
                        "Invalid acquire mode! The live hardware doesn't support "
                        "RAW acquire mode!"
                    )

            # Check if we've got a measure in the middle of the circuit somewhere.
            elif isinstance(inst, Acquire):
                for qbit in model.qubits:
                    if qbit.get_measure_channel() == inst.channel:
                        consumed_qubits.append(qbit)
            elif isinstance(inst, Pulse):
                # Find target qubit from instruction and check whether it's been
                # measured already.
                acquired_qubits = [
                    model._resolve_qb_pulse_channel(chanbit)[0] in consumed_qubits
                    for chanbit in inst.quantum_targets
                    if isinstance(chanbit, (Qubit, PulseChannel))
                ]

                if any(acquired_qubits):
                    raise ValueError(
                        "Mid-circuit measurements currently unable to be used."
                    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from __future__ import annotations

from typing import Any, List, Set

from qat.purr.compiler.instructions import (
    Acquire,
    PostProcessing,
    QuantumInstruction,
    Repeat,
    Return,
    Sweep,
)


class TimelineSegment:
    """Segment of the execution timeline."""

    def __init__(
        self,
        instruction: QuantumInstruction,
        dependencies: Set[str] = None,
        reliant: List[Any] = None,
    ):
        self.instruction = instruction
        self.scheduling_dependencies = dependencies or {
            comp.full_id() for comp in instruction.quantum_targets
        }
        self.reliant_instructions = reliant or []

    def __repr__(self):
        sep = ";" if any(self.reliant_instructions) else ""
        sched_deps = (
            f"[{','.join(self.scheduling_dependencies)}]"
            if any(self.scheduling_dependencies)
            else ""
        )
        return (
            f"{self.instruction.duration}: {sched_deps} {str(self.instruction)}"
            f"{sep}{';'.join(str(inst) for inst in self.reliant_instructions)}"
        )


class QatFile:
    """In-memory representation of our instruction file."""

    def __init__(self):
        self.timeline: List[TimelineSegment] = []
        self.meta_instructions = []

    def add(self, *args, **kwargs):
        self.timeline.append(TimelineSegment(*args, **kwargs))

    def add_meta(self, instruction):
        if isinstance(instruction, Return):
            existing_return = next(
                iter(meta for meta in self.meta_instructions if isinstance(meta, Return)),
                None,
            )
            if existing_return is not None:
                existing_return.variables.extend(instruction.variables)
            else:
                self.meta_instructions.append(instruction)
        else:
            self.meta_instructions.append(instruction)

    @property
    def instructions(self):
        return [seg.instruction for seg in self.timeline]

    def get_pp_for_variable(self, target_var):
        results = []
        for instruction in self.instructions:
            if (
                isinstance(instruction, PostProcessing)
                and instruction.output_variable == target_var
            ):
                results.append(instruction)

        return results

    @property
    def sweeps(self):
        return [sw for sw in self.meta_instructions if isinstance(sw, Sweep)]

    @property
    def repeat(self):
        return next(
            iter(sw for sw in self.meta_instructions if isinstance(sw, Repeat)), None
        )

    @property
    def return_(self):
        return next(
            iter(sw for sw in self.meta_instructions if isinstance(sw, Return)), None
        )


class InstructionEmitter:
    """
    Interface to the LLVM-driven QAT optimization and construction passes. For now we
    simulate what it might do in the future and ust output a Python object that
    simulates what our instruction set might look like.
    """

    def emit(self, instructions, hardware):
        qatf = QatFile()

        for inst in instructions:
            if isinstance(inst, PostProcessing):
                qatf.add(inst, [inst.acquire.channel.full_id()])
            elif isinstance(inst, QuantumInstruction):
                qatf.add(inst)
            else:
                qatf.add_meta(inst)

        # If we don't have an explicit return, imply all results.
        if not any(ret for ret in qatf.meta_instructions if isinstance(ret, Return)):
            # Only gather each variable once for the return.
            unique_variables = []
            for var in [
                acq.output_variable for acq in qatf.instructions if isinstance(acq, Acquire)
            ]:
                if var not in unique_variables:
                    unique_variables.append(var)
            qatf.add_meta(Return(unique_variables))

        # If we don't have a repeat, set it as 1.
        repeat_inst = next(
            (rep for rep in qatf.meta_instructions if isinstance(rep, Repeat)), None
        )
        if repeat_inst is None:
            qatf.add_meta(
                Repeat(hardware.default_repeat_count, hardware.default_repetition_period)
            )
        else:
            if repeat_inst.repeat_count is None:
                repeat_inst.repeat_count = hardware.default_repeat_count
            if repeat_inst.repetition_period is None:
                repeat_inst.repetition_period = hardware.default_repetition_period

        return qatf



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from collections.abc import Iterable
from numbers import Number
from typing import List, Optional, TypeVar, Union

import numpy
from compiler_config.config import (
    CalibrationArguments,
    CompilerConfig,
    ErrorMitigationConfig,
    MetricsType,
    ResultsFormatting,
)

from qat.compiler.transform_passes import PhaseOptimisation, PostProcessingOptimisation
from qat.compiler.validation_passes import InstructionValidation, ReadoutValidation
from qat.ir.pass_base import InvokerMixin, PassManager
from qat.ir.result_base import ResultManager
from qat.purr.compiler.builders import (
    FluidBuilderWrapper,
    InstructionBuilder,
    QuantumInstructionBuilder,
)
from qat.purr.compiler.error_mitigation.readout_mitigation import get_readout_mitigation
from qat.purr.compiler.execution import (
    InstructionExecutionEngine,
    QuantumExecutionEngine,
    _binary,
)
from qat.purr.compiler.hardware_models import QuantumHardwareModel, get_cl2qu_index_mapping
from qat.purr.compiler.instructions import Instruction, is_generated_name
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.compiler.metrics import CompilationMetrics, MetricsMixin
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class RemoteCalibration:
    """
    Base class for any remote calibration executions. These are far more complicated
    blocks than purely a string of instructions and include nested executions and rely
    on classic Python code.
    """

    def run(
        self,
        model: QuantumHardwareModel,
        runtime: "QuantumRuntime",
        args: CalibrationArguments,
    ):
        raise ValueError("Calibration cannot be run at this time.")

    def arguments_type(self) -> type:
        """Returns the type of this calibrations arguments."""
        return CalibrationArguments


class QuantumExecutableBlock:
    """Generic executable block that can be run on a quantum runtime."""

    def run(self, runtime: "QuantumRuntime"):
        pass


class CalibrationWithArgs(QuantumExecutableBlock):
    """Wrapper for a calibration and argument combination."""

    def __init__(self, calibration: RemoteCalibration, args: CalibrationArguments = None):
        self.calibration = calibration
        self.args = args or CalibrationArguments()

    def run(self, runtime: "QuantumRuntime"):
        if self.calibration is None:
            raise ValueError("No calibration to run.")

        self.calibration.run(runtime.model, runtime, self.args)


AnyEngine = TypeVar("AnyEngine", bound=InstructionExecutionEngine, covariant=True)


class QuantumRuntime(MetricsMixin):
    def __init__(self, execution_engine: InstructionExecutionEngine, metrics=None):
        super().__init__()
        self.engine: AnyEngine = execution_engine
        self.compilation_metrics = metrics or CompilationMetrics()

    @property
    def model(self):
        return self.engine.model if self.engine is not None else None

    def _transform_results(
        self, results, format_flags: ResultsFormatting, repeats: Optional[int] = None
    ):
        """
        Transform the raw results into the format that we've been asked to provide. Look
        at individual transformation documentation for descriptions on what they do.
        """
        if len(results) == 0:
            return []

        # If we have no flags at all just infer structure simplification.
        if format_flags is None:
            format_flags = ResultsFormatting.DynamicStructureReturn

        if repeats is None:
            repeats = 1000

        def simplify_results(simplify_target):
            """
            To facilitate backwards compatability and being able to run low-level
            experiments alongside quantum programs we make some assumptions based upon
            form of the results.

            If all results have default variable names then the user didn't care about
            value assignment or this was a low-level experiment - in both cases, it
            means we can throw away the names and simply return the results in the order
            they were defined in the instructions.

            If we only have one result after this, just return that list directly
            instead, as it's probably just a single experiment.
            """
            if all([is_generated_name(k) for k in simplify_target.keys()]):
                if len(simplify_target) == 1:
                    return list(simplify_target.values())[0]
                else:
                    squashed_results = list(simplify_target.values())
                    if all(isinstance(val, numpy.ndarray) for val in squashed_results):
                        return numpy.array(squashed_results)
                    return squashed_results
            else:
                return simplify_target

        if ResultsFormatting.BinaryCount in format_flags:
            results = {key: _binary_count(val, repeats) for key, val in results.items()}

        def squash_binary(value):
            if isinstance(value, int):
                return str(value)
            elif all(isinstance(val, int) for val in value):
                return "".join([str(val) for val in value])

        if ResultsFormatting.SquashBinaryResultArrays in format_flags:
            results = {key: squash_binary(val) for key, val in results.items()}

        # Dynamic structure return is an ease-of-use flag to strip things that you know
        # your use-case won't use, such as variable names and nested lists.
        if ResultsFormatting.DynamicStructureReturn in format_flags:
            results = simplify_results(results)

        return results

    def _apply_error_mitigation(self, results, instructions, error_mitigation):
        if error_mitigation is None or error_mitigation == ErrorMitigationConfig.Empty:
            return results

        # TODO: add support for multiple registers
        # TODO: reconsider results length
        if len(results) > 1:
            raise ValueError(
                "Cannot have multiple registers in conjunction with readout error mitigation."
            )

        mapping = get_cl2qu_index_mapping(instructions, self.model)
        for mitigator in get_readout_mitigation(error_mitigation):
            new_result = mitigator.apply_error_mitigation(results, mapping, self.model)
            results[mitigator.name] = new_result
        return results  # TODO: new results object

    def run_calibration(
        self, calibrations: Union[CalibrationWithArgs, List[CalibrationWithArgs]]
    ):
        """Make 'calibration' distinct from 'quantum executable' for usabilities sake."""
        self.run_quantum_executable(calibrations)

    def run_quantum_executable(
        self, executables: Union[QuantumExecutableBlock, List[QuantumExecutableBlock]]
    ):
        if executables is None:
            return

        if not isinstance(executables, list):
            executables = [executables]

        for exe in executables:
            exe.run(self)

    def _common_execute(
        self,
        fexecute: callable,
        instructions,
        results_format=None,
        repeats=None,
        error_mitigation=None,
    ):
        """
        Executes these instructions against the current engine and returns the results.
        """
        if self.engine is None:
            raise ValueError("No execution engine available.")

        if isinstance(instructions, InstructionBuilder) or isinstance(
            instructions, FluidBuilderWrapper
        ):
            instructions = instructions.instructions

        if instructions is None or not any(instructions):
            raise ValueError(
                "No instructions passed to the process or stored for execution."
            )

        instructions = self.engine.optimize(instructions)
        self.engine.validate(instructions)
        self.record_metric(
            MetricsType.OptimizedInstructionCount, opt_inst_count := len(instructions)
        )
        log.info(f"Optimized instruction count: {opt_inst_count}")

        results = fexecute(instructions)
        results = self._transform_results(results, results_format, repeats)
        return self._apply_error_mitigation(results, instructions, error_mitigation)

    def _execute_with_interrupt(
        self,
        instructions,
        results_format=None,
        repeats=None,
        interrupt: Interrupt = NullInterrupt(),
        error_mitigation=None,
    ):
        """
        Executes these instructions against the current engine and returns the results.
        """

        def fexecute(instrs):
            return self.engine._execute_with_interrupt(instrs, interrupt)

        return self._common_execute(
            fexecute, instructions, results_format, repeats, error_mitigation
        )

    def execute(
        self, instructions, results_format=None, repeats=None, error_mitigation=None
    ):
        """
        Executes these instructions against the current engine and returns the results.
        """

        def fexecute(instrs):
            return self.engine.execute(instrs)

        return self._common_execute(
            fexecute, instructions, results_format, repeats, error_mitigation
        )


class NewQuantumRuntime(QuantumRuntime, InvokerMixin):
    """
    Uses the new pass infrastructure.

    Notice how polymorphic calls to XEngine.optimize() and XEngine.validate() are avoided. Instead, we have
    a flat structure of passes. This allows developers to focus on efficiently implementing a pass and easily test,
    demonstrate, and register passes without worrying too much about where it fits into the global compilation
    workflow.

    The NewQuantumRuntime deliberately recognises the builder as the only acceptable form of input "IR" and refuses
    to take in a bare list of instructions. This reduces the constant confusion of "builder" vs "instructions".

    The NewQuantumRuntime is also deliberately stripped out of any handling of compilation metrics. In fact, ideas
    similar to the new pass infrastructure can be applied to compilation metrics, that's why we're excluding them
    during this iteration partly because other pieces need to come together and partly because the current iteration
    needs to be kept light-weight and technically tractable.
    """

    def build_pass_pipeline(self, *args, **kwargs):
        return (
            PassManager()
            | PhaseOptimisation()
            | PostProcessingOptimisation()
            | InstructionValidation()
            | ReadoutValidation()
        )

    def _common_execute(
        self,
        fexecute: callable,
        builder: InstructionBuilder,
        results_format=None,
        repeats=None,
        error_mitigation=None,
    ):
        if self.engine is None:
            raise ValueError("No execution engine available.")

        if not isinstance(builder, InstructionBuilder):
            raise ValueError(
                f"Expected InstructionBuilder, but got {type(builder)} instead"
            )

        res_mgr = ResultManager()
        self.run_pass_pipeline(builder, res_mgr, self.model, self.engine)
        results = fexecute(builder)
        results = self._transform_results(results, results_format, repeats)
        return self._apply_error_mitigation(results, builder, error_mitigation)


def _binary_count(results_list, repeats):
    """
    Returns a dictionary of binary number: count. So for a two qubit register it'll return the various counts for
    ``00``, ``01``, ``10`` and ``11``.
    """

    def flatten(res):
        """
        Combine binary result from the QPU into composite key result.
        Aka '0110' or '0001'
        """
        if isinstance(res, Iterable):
            return "".join([flatten(val) for val in res])
        else:
            return str(res)

    def get_tuple(res, index):
        return [
            val[index] if isinstance(val, (List, numpy.ndarray)) else val for val in res
        ]

    binary_results = _binary(results_list)

    # If our results are a single qubit then pretend to be a register of one.
    if (
        isinstance(next(iter(binary_results), None), Number)
        and len(binary_results) == repeats
    ):
        binary_results = [binary_results]

    result_count = dict()
    for qubit_result in [list(get_tuple(binary_results, i)) for i in range(repeats)]:
        key = flatten(qubit_result)
        value = result_count.get(key, 0)
        result_count[key] = value + 1

    return result_count


def get_model(hardware: Union[QuantumExecutionEngine, QuantumHardwareModel]):
    if isinstance(hardware, QuantumExecutionEngine):
        return hardware.model
    return hardware


def get_runtime(hardware: Union[QuantumExecutionEngine, QuantumHardwareModel]):
    # We shouldn't really have an orphaned execution engine, but no harm in it.
    if isinstance(hardware, QuantumExecutionEngine):
        return hardware.model.create_runtime(hardware)
    elif isinstance(hardware, QuantumHardwareModel):
        return hardware.create_runtime()

    raise ValueError(
        f"{str(hardware)} is not a recognized hardware model or execution engine."
    )


def get_builder(
    model: Union[QuantumHardwareModel, InstructionExecutionEngine]
) -> QuantumInstructionBuilder:
    if isinstance(model, InstructionExecutionEngine):
        model = model.model
    return model.create_builder()


def execute_instructions(
    hardware: Union[QuantumExecutionEngine, QuantumHardwareModel],
    instructions: Union[List[Instruction], QuantumInstructionBuilder],
    config: CompilerConfig = None,
    executable_blocks: List[QuantumExecutableBlock] = None,
    repeats: Optional[int] = None,
    metrics: MetricsType = None,
    *args,
    **kwargs,
):
    if config is None:
        config = CompilerConfig()
    config.validate(hardware)

    active_runtime = get_runtime(hardware)

    active_runtime.run_quantum_executable(executable_blocks)
    results = active_runtime.execute(
        instructions,
        config.results_format,
        config.repeats,
        config.error_mitigation,
    )
    return results, active_runtime.compilation_metrics


def _execute_instructions_with_interrupt(
    hardware: Union[QuantumExecutionEngine, QuantumHardwareModel],
    instructions: Union[List[Instruction], QuantumInstructionBuilder],
    config: CompilerConfig = None,
    executable_blocks: List[QuantumExecutableBlock] = None,
    interrupt: Interrupt = NullInterrupt(),
    *args,
    **kwargs,
):
    if config is None:
        config = CompilerConfig()
    config.validate(hardware)

    active_runtime = get_runtime(hardware)
    active_runtime.run_quantum_executable(executable_blocks)

    results = active_runtime._execute_with_interrupt(
        instructions,
        config.results_format,
        config.repeats,
        interrupt,
        config.error_mitigation,
    )
    return results, active_runtime.compilation_metrics



import uuid
from abc import ABC
from typing import Set

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class ResultConcept(ABC):
    """
    Base class describing the abstraction of an analysis result.

    See ResultManager.
    """

    pass


class ResultModel(ResultConcept):
    """
    Wrapper for any result object typically produced by an analysis pass.

    See ResultManager.
    """

    def __init__(self, res_obj):
        self._result = res_obj
        self._uuid = uuid.uuid4()

    @property
    def value(self):
        return self._result

    def __hash__(self):
        return hash(self._uuid)


class ResultInfoMixin(ABC):
    """
    Base mixin specifying result identification mechanism. A result has an id, name, and value.

    See ResultManager.
    """

    def id(self):
        pass

    def name(self):
        pass

    def value(self):
        pass


class ResultManager:
    """
    Represents a collection of analysis results with caching and aggregation capabilities.

    Passes that merely compute analyses on the IR must not invalidate prior results. Passes that mutate
    any IR units are likely to invalidate predecessor results.

    An analysis pass can produce 1 or more result objects. There is in theory a duality between passes
    and the results they produce.

    To keep things simple, the ResultManager is just a set of analysis results. Result identification is also
    kept trivial where a UUID is used internally. Proper identification mechanism will be called upon
    once we feel the need for a more sophisticated PassManager.

    See PassManager.
    """

    def __init__(self):
        self._results: Set[ResultModel] = set()

    @property
    def results(self):
        return self._results

    def update(self, other_res_mgr):
        if not isinstance(other_res_mgr, ResultManager):
            raise ValueError(
                f"Invalid type, expected {ResultManager}, but got {type(other_res_mgr)}"
            )
        self._results.update(other_res_mgr._results)

    def add(self, res_obj):
        self._results.add(ResultModel(res_obj))

    def lookup_by_type(self, ty: type):
        found = [res.value for res in self._results if isinstance(res.value, ty)]
        if not found:
            raise ValueError(f"Could not find any results instances of {ty}")

        if len(found) > 1:
            raise ValueError(f"Found multiple results instances of {ty}")

        return found[0]
